//
// Lighthouse3D.com OpenGL Texturing Sample
//
// Uses:
//  Devil for image loading
//		http://openil.sourceforge.net/
//
// The texture was downloaded from:
// http://opengameart.org/content/basket-ball-texture
//
// This demo was built for learning purposes only.
// Some code could be severely optimised, but I tried to
// keep as simple and clear as possible.
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//
// If you do use it I would love to hear about it. Just post a comment
// at Lighthouse3D.com

// Have Fun :-)

//
// Program for lecturing purposes by Marek Vinkler (xvinkl@fi.muni.cz).
//
// The code comes with no warranties, use it at your own risk.
// You may use it, or parts of it, wherever you want.
//


#include <cmath>

// Include GLEW to access OpenGL 3.3 functions
#include <GL/glew.h>

// GLUT is the toolkit to interface with the OS
#include <GL/freeglut.h>

// Use Very Simple Libs
#include "vsMathLib.h"
#include "vsShaderLib.h"
#include "vsResSurfRevLib.h"

//----------------------------------------------------------------
// TODO - _DIFFUSE:
// 1) Inspect the texture initialization in function "init".

// TODO - _DISSOLVE:
// 1) Use the histogram corrected image, explain the difference.

// TODO - _CUBEMAP:
// 1) See how two shaders are initialized in the "setupShaders" function.
// 2) Notice how cube map geometry and textures are initialized in "init".
// 3) Inspect the skybox shaders.
// 4) Render the cube map geometry with the corresponding shader.

// TODO - _NORMALMAP:
// 1) Notice another texture is loaded in function "init".
// 2) Try different normal maps.

// TODO - _NORMALMAPTANGENT:
// 1) Notice the addition of tangent attributes in function "setupShaders".
// 2) Increase the material's specular component to visualize the difference between the methods.
// 3) Optionally check the code in VSResSurfRevLib which generates tangent vectors.
//----------------------------------------------------------------

#define _DIFFUSE			0
#define _DISSOLVE			1
#define _CUBEMAP			2
#define _NORMALMAP			3 // Object space normal map - not portable to other orientations
#define _NORMALMAPTANGENT	4 // Object space normal map - portable to other orientations

#define _TEXTURE _DIFFUSE
//----------------------------------------------------------------

//#define INV_VIEW_NORMAL

VSMathLib *vsml;
VSShaderLib shader, shaderSkybox;
VSResSurfRevLib surfRev, surfRevSkybox;

// Camera Position
float camX, camY, camZ;

// Mouse Tracking Variables
int startX, startY, tracking = 0;

// Camera Spherical Coordinates
float alpha = 0.0f, beta = 0.0f;
float r = 5.0f;

// Geometry offset
float offsetX = 0.f, offsetY = 0.f;

float time = 0.f;


// ------------------------------------------------------------
//
// Reshape Callback Function
//

void changeSize(int w, int h)
{
	float ratio;

	// Prevent a divide by zero, when window is too short
	if(h == 0)
		h = 1;

	// Set the viewport to be the entire window
    glViewport(0, 0, w, h);

	ratio = (1.0f * w) / h;
	vsml->loadIdentity(VSMathLib::PROJECTION);
	vsml->perspective(53.13f, ratio, 0.1f, 10000.0f);
}


// ------------------------------------------------------------
//
// Render stuff
//

float lightDir[4] = {-1.0f, 0.0f, 1.0f, 0.0f};

void renderScene(void)
{
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	vsml->loadIdentity(VSMathLib::VIEW);
	vsml->loadIdentity(VSMathLib::MODEL);
	
	// Set camera
	vsml->lookAt(camX, camY, camZ, 0,0,0, 0,1,0);
  
	vsml->translate(offsetX, offsetY, 0.f);

#if (_TEXTURE == _CUBEMAP)
	// Use skybox shader
	glUseProgram(shaderSkybox.getProgramIndex());

	surfRevSkybox.render();
#endif

	// Transform light to camera space and send it to GLSL
	float res[4];
	vsml->multMatrixPoint(VSMathLib::VIEW, lightDir, res);
	vsml->normalize(res);
	shader.setBlockUniform("Lights", "l_dir", res);

	// Use our shader
	glUseProgram(shader.getProgramIndex());

	shader.setUniform("time", time);

	surfRev.render();

	time += 0.01f;

	// Swap buffers
	glutSwapBuffers();
}


// ------------------------------------------------------------
//
// Events from the Keyboard
//

void processKeys(unsigned char key, int xx, int yy)
{
	switch(key)
	{
		case 27:
			glutLeaveMainLoop();
			break;

		case 'c': printf("Camera Spherical Coordinates (%f, %f, %f)\n", alpha, beta, r);
			break;

		case 'm': glEnable(GL_MULTISAMPLE);
			break;

		case 'n': glDisable(GL_MULTISAMPLE);
			 break;
	}

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// ------------------------------------------------------------
//
// Mouse Events
//

void processMouseButtons(int button, int state, int xx, int yy)
{
	// Start tracking the mouse
	if (state == GLUT_DOWN)
	{
		startX = xx;
		startY = yy;
		if (button == GLUT_LEFT_BUTTON)
			tracking = 1;
		else if (button == GLUT_RIGHT_BUTTON)
			tracking = 2;
		else if (button == GLUT_MIDDLE_BUTTON)
			tracking = 3;
	}

	// Stop tracking the mouse
	else if (state == GLUT_UP)
	{
		if (tracking == 1)
		{
			alpha -= (xx - startX);
			beta += (yy - startY);
			if (beta > 85.0f)
				beta = 85.0f;
			else if (beta < -85.0f)
				beta = -85.0f;
		}
		else if (tracking == 2)
		{
			r += (yy - startY) * 0.01f;
			if (r < 0.1f)
				r = 0.1f;
		}
		tracking = 0;
	}
}

// Track mouse motion while buttons are pressed
void processMouseMotion(int xx, int yy)
{
	int deltaX, deltaY;
	float alphaAux, betaAux;
	float rAux;

	deltaX =  - xx + startX;
	deltaY =    yy - startY;

	// Left mouse button: move camera
	if (tracking == 1)
	{
		alphaAux = alpha + deltaX;
		betaAux = beta + deltaY;

		if (betaAux > 85.0f)
			betaAux = 85.0f;
		else if (betaAux < -85.0f)
			betaAux = -85.0f;
		rAux = r;
	}
	// Right mouse button: zoom
	else if (tracking == 2)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r + (deltaY * 0.01f);
		if (rAux < 0.1f)
			rAux = 0.1f;
	}
	// Middle mouse button: move object
	else if (tracking == 3)
	{
		alphaAux = alpha;
		betaAux = beta;
		rAux = r;

		offsetX -= deltaX * 0.0001f;
		offsetY -= deltaY * 0.0001f;
	}

	camX = rAux * sin(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camZ = rAux * cos(alphaAux * 3.14f / 180.0f) * cos(betaAux * 3.14f / 180.0f);
	camY = rAux *   						       sin(betaAux * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}

void mouseWheel(int wheel, int direction, int x, int y)
{
	r += direction * 0.1f;
	if (r < 0.1f)
		r = 0.1f;

	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

//  uncomment this if not using an idle func
//	glutPostRedisplay();
}


// --------------------------------------------------------
//
// Shader Stuff
//

bool setupShaders()
{
	// Shader for models
	shader.init();

#if (_TEXTURE == _DIFFUSE)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pixeldirdifambspec.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/pixeldirdifambspec.frag");

#elif (_TEXTURE == _DISSOLVE)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pixeldirdifambspec.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/dissolve.frag");

#elif (_TEXTURE == _CUBEMAP)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/cubemap.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/cubemap.frag");
	
	// Shader for the skybox
	shaderSkybox.init();

	shaderSkybox.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/skybox.vert");
	shaderSkybox.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/skybox.frag");

	// Set semantics for the shader variables
	shaderSkybox.setProgramOutput(0, "outputF");
	shaderSkybox.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");


	printf("InfoLog\n%s\n\n", shaderSkybox.getAllInfoLogs().c_str());
	shaderSkybox.prepareProgram();

	// Add sampler uniform
	shaderSkybox.setUniform("cubemap", 0);

#elif (_TEXTURE == _NORMALMAP)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/pixeldirdifambspec.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/normalmap.frag");

#elif (_TEXTURE == _NORMALMAPTANGENT)
	shader.loadShader(VSShaderLib::VERTEX_SHADER, "shaders/normalmaptangent.vert");
	shader.loadShader(VSShaderLib::FRAGMENT_SHADER, "shaders/normalmaptangent.frag");
#endif

	// Set semantics for the shader variables
	shader.setProgramOutput(0, "outputF");
	shader.setVertexAttribName(VSShaderLib::VERTEX_COORD_ATTRIB, "position");
	shader.setVertexAttribName(VSShaderLib::NORMAL_ATTRIB, "normal");
	shader.setVertexAttribName(VSShaderLib::TEXTURE_COORD_ATTRIB, "texCoord");
#if (_TEXTURE == _NORMALMAPTANGENT)
	shader.setVertexAttribName(VSShaderLib::TANGENT_ATTRIB, "tangent");
	shader.setVertexAttribName(VSShaderLib::BITANGENT_ATTRIB, "bitangent");
#endif


	printf("InfoLog\n%s\n\n", shader.getAllInfoLogs().c_str());
	shader.prepareProgram();

	// Add sampler uniform
	shader.setUniform("texUnit", 0);

#if (_TEXTURE == _NORMALMAP) || (_TEXTURE == _NORMALMAPTANGENT)
	shader.setUniform("normalUnit", 1);
#endif

	
	return(shader.isProgramValid());
}


// ------------------------------------------------------------
//
// Model loading and OpenGL setup
//

bool init()
{
	camX = r * sin(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camZ = r * cos(alpha * 3.14f / 180.0f) * cos(beta * 3.14f / 180.0f);
	camY = r *   						     sin(beta * 3.14f / 180.0f);

	surfRev.createSphere(1.f, 32.f);

	surfRev.setMaterialBlockName("Materials");

#if (_TEXTURE == _DIFFUSE)
	surfRev.addTexture(0, "textures/balldimpled.png");

	float f[4] = {1.0f, 1.0f, 1.0f, 1.0f};
	surfRev.setColor(VSResourceLib::DIFFUSE, f);
	float f2[4] = {0.2f, 0.2f, 0.2f, 1.0f};
	surfRev.setColor(VSResourceLib::AMBIENT, f2);

#elif (_TEXTURE == _DISSOLVE)
	surfRev.addTexture(0, "textures/clouds.png");
	//surfRev.addTexture(0, "textures/clouds_hist.png");

	float f[4] = {0.4f, 0.6f, 1.0f, 1.0f};
	surfRev.setColor(VSResourceLib::DIFFUSE, f);
	float f2[4] = {0.1f, 0.15f, 0.25f, 1.0f};
	surfRev.setColor(VSResourceLib::AMBIENT, f2);

#elif (_TEXTURE == _CUBEMAP)
	surfRev.addCubeMapTexture(0, "textures/skybox/xpos.png", "textures/skybox/xneg.png", "textures/skybox/ypos.png", "textures/skybox/yneg.png", "textures/skybox/zpos.png", "textures/skybox/zneg.png");

	surfRevSkybox.createSphere(20.f, 32.f);
	surfRevSkybox.addCubeMapTexture(0, "textures/skybox/xpos.png", "textures/skybox/xneg.png", "textures/skybox/ypos.png", "textures/skybox/yneg.png", "textures/skybox/zpos.png", "textures/skybox/zneg.png");

#elif (_TEXTURE == _NORMALMAP) || (_TEXTURE == _NORMALMAPTANGENT)
	surfRev.addTexture(0, "textures/Shaft-1.jpg");
	surfRev.addTexture(1, "textures/normalmap/BrickNormal.jpg");
	//surfRev.addTexture(1, "textures/normalmap/snakebump.jpg");

	// Object space / tangent space normal maps. The difference would be easier to spot on a plane.
	//surfRev.addTexture(1, "textures/normalmap/04_face.png");
	//surfRev.addTexture(1, "textures/normalmap/16_face.png");

	float f[4] = {0.4f, 0.6f, 1.0f, 1.0f};
	surfRev.setColor(VSResourceLib::DIFFUSE, f);
	float f2[4] = {0.1f, 0.15f, 0.25f, 1.0f};
	surfRev.setColor(VSResourceLib::AMBIENT, f2);
	float f3[4] = {0.2f, 0.2f, 0.2f, 1.0f};
	surfRev.setColor(VSResourceLib::SPECULAR, f3);
#endif

	// Some GL settings
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_MULTISAMPLE);

	glClearColor(1.0f, 1.0f, 1.0f, 1.0f);

	return true;
}

void initVSL()
{
//	Init VSML
	vsml = VSMathLib::getInstance();
	vsml->setUniformBlockName("Matrices");
	vsml->setUniformName(VSMathLib::PROJ_VIEW_MODEL, "m_pvm");
	vsml->setUniformName(VSMathLib::NORMAL, "m_normal");
	vsml->setUniformName(VSMathLib::VIEW_MODEL, "m_viewModel");
#ifndef INV_VIEW_NORMAL
	vsml->setUniformName(VSMathLib::VIEW, "m_view");
#else
	vsml->setUniformName(VSMathLib::NORMAL_VIEW, "m_view");
#endif
}


// ------------------------------------------------------------
//
// Main function
//

int main(int argc, char **argv)
{
//  GLUT initialization
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DEPTH|GLUT_DOUBLE|GLUT_RGBA|GLUT_MULTISAMPLE);

	glutInitContextVersion(3, 3);
	glutInitContextProfile(GLUT_CORE_PROFILE);
	glutInitContextFlags(GLUT_DEBUG);

	glutInitWindowPosition(100,100);
	glutInitWindowSize(512,512);
	glutCreateWindow("PV227 - Texture Maps");


//  Callback Registration
	glutDisplayFunc(renderScene);
	glutReshapeFunc(changeSize);
	glutIdleFunc(renderScene);

//	Mouse and Keyboard Callbacks
	glutKeyboardFunc(processKeys);
	glutMouseFunc(processMouseButtons);
	glutMotionFunc(processMouseMotion);

	glutMouseWheelFunc(mouseWheel);

//	Return from main loop
	glutSetOption(GLUT_ACTION_ON_WINDOW_CLOSE, GLUT_ACTION_GLUTMAINLOOP_RETURNS);

//	Init GLEW
	glewExperimental = GL_TRUE;
	glewInit();

	printf("Vendor: %s\n", glGetString(GL_VENDOR));
	printf("Renderer: %s\n", glGetString(GL_RENDERER));
	printf("Version: %s\n", glGetString(GL_VERSION));
	printf("GLSL: %s\n", glGetString(GL_SHADING_LANGUAGE_VERSION));
	int param;
	glGetIntegerv(GL_CONTEXT_PROFILE_MASK, &param);
	if(param == GL_CONTEXT_CORE_PROFILE_BIT)
		printf("Context Profile: Core\n");
	else
		printf("Context Profile: Compatibility\n");

	if(!setupShaders())
		exit(1);

	if(!init())
		exit(1);

	initVSL();

//  GLUT main loop
	glutMainLoop();

	return(0);
}

